from datetime import datetime, timezone
from random import randint
from typing import Optional

from faker import Faker

from t_reporting_sdk.domain.paula_records.models import PAULARecord, PAULARecordStatus, PAULARecordOriginalStatus, PAULAExceptionType


class PAULARecordFaker:
    @staticmethod
    def provide(
        status: Optional[PAULARecordStatus] = None,
        exception_type: Optional[PAULAExceptionType] = None,
        message: Optional[str] = None,
        customer_id: Optional[str] = None,
        original_status: Optional[PAULARecordOriginalStatus] = None,
        patient_id: Optional[str] = None,
        portal: Optional[str] = None,
        appointment_date: Optional[datetime] = None,
        treatment_code: Optional[str] = None,
        payer: Optional[str] = None,
        payer_id: Optional[str] = None,
        plan_id: Optional[str] = None,
        system_of_record_id: Optional[str] = None,
        authorization_id: Optional[str] = None,
    ) -> PAULARecord:
        faker = Faker()

        fake_status = status or faker.enum(PAULARecordStatus)
        fake_exception_type = exception_type or (faker.enum(PAULAExceptionType) if fake_status == PAULARecordStatus.ERROR else None)
        fake_message = message or (faker.text(max_nb_chars=100) if fake_exception_type is not None else None)
        fake_customer_id = customer_id or str(randint(1, 100))
        fake_original_status = original_status or faker.enum(PAULARecordOriginalStatus)
        fake_patient_id = patient_id or str(randint(1, 100))
        fake_portal = portal or faker.company()
        fake_appointment_date = appointment_date or faker.date_time_this_month(tzinfo=timezone.utc)
        fake_treatment_code = treatment_code or faker.word()
        fake_payer = payer or (faker.company() if fake_status in [PAULARecordStatus.PENDING, PAULARecordStatus.DENIED, PAULARecordStatus.APPROVED] else None)
        fake_payer_id = payer_id or (str(randint(1, 100)) if fake_status in [PAULARecordStatus.PENDING, PAULARecordStatus.DENIED, PAULARecordStatus.APPROVED] else None)
        fake_plan_id = plan_id or (str(randint(1, 100)) if fake_payer is not None else None)
        fake_system_of_record_id = system_of_record_id or str(randint(1, 100))
        fake_authorization_id = authorization_id or (str(randint(1, 100)) if fake_status in [PAULARecordStatus.DENIED, PAULARecordStatus.APPROVED] else None)

        return PAULARecord(
            status=fake_status,
            exception_type=fake_exception_type,
            message=fake_message,
            customer_id=fake_customer_id,
            original_status=fake_original_status,
            patient_id=fake_patient_id,
            portal=fake_portal,
            appointment_date=fake_appointment_date,
            treatment_code=fake_treatment_code,
            payer=fake_payer,
            payer_id=fake_payer_id,
            plan_id=fake_plan_id,
            system_of_record_id=fake_system_of_record_id,
            authorization_id=fake_authorization_id,
        )


from typing import Optional, List
import logging
from concurrent.futures import ThreadPoolExecutor, as_completed

from t_reporting_sdk.domain.agent_runs.models import AgentRun
from t_reporting_sdk.domain.cam_records.models import CAMRecord
from t_reporting_sdk.repositories.agent_runs.repository import AgentRunsRepository
from t_reporting_sdk.repositories.cam_records.repository import CAMRecordsRepository


def _load_default_cam_records_repository() -> CAMRecordsRepository:
    """
    Handles dependency injection automatically so that SDK users do not have to specify their dependencies manually.
    Handled as a separate function so that tests do not try to import the default repositories
    """
    from t_reporting_sdk.default_dependencies import default_cam_records_repository
    return default_cam_records_repository


def _load_default_agent_runs_repository() -> AgentRunsRepository:
    """
    Handles dependency injection automatically so that SDK users do not have to specify their dependencies manually.
    Handled as a separate function so that tests do not try to import the default repositories
    """
    from t_reporting_sdk.default_dependencies import default_agent_runs_repository
    return default_agent_runs_repository


class CAMReporter:
    def __init__(
            self,
            agent_run: AgentRun,
            cam_repository: Optional[CAMRecordsRepository] = None,
            agent_runs_repository: Optional[AgentRunsRepository] = None,
    ):
        # Load default repositories if none are provided
        self._cam_repository = _load_default_cam_records_repository() if cam_repository is None else cam_repository
        agent_runs_repository = _load_default_agent_runs_repository() if agent_runs_repository is None else agent_runs_repository

        self._agent_run = agent_run

        # Ensure Agent Run is stored in the database so that we can associate CAM results with it
        agent_runs_repository.store_agent_run(agent_run=agent_run)

    def report_cam_records(
            self,
            cam_records: List[CAMRecord],
            # Limit the # of saves made parallel - saves will fail if # of parallel calls > connection_pool size
            max_records_to_store_in_parallel: int = 4,
    ) -> None:
        def store_record(cam_record: CAMRecord) -> None:
            try:
                self._cam_repository.store_cam_record(
                    agent_run=self._agent_run,
                    cam_record=cam_record,
                )
            except Exception as e:
                logging.error(f"Unable to report CAM result: {e}")

        with ThreadPoolExecutor(max_workers=max_records_to_store_in_parallel) as executor:
            futures = {executor.submit(store_record, record): record for record in cam_records}

            for future in as_completed(futures):
                record = futures[future]
                try:
                    future.result()
                except Exception as e:
                    logging.error(f"Failed to process record {record}: {e}")


from enum import Enum
from typing import Optional

from pydantic import BaseModel


class CAMRecordStatus(Enum):
    # Successfully processed entire claim. No issues/warnings encountered.
    SUCCESS = 'success'

    # Successfully processed claim but encountered a scenario that requires staff review.
    WARNING = 'warning'

    # Unable to process claim
    ERROR = 'error'

    # Skipped typically indicates that the scenario is a known-out of scope scenario
    SKIPPED = 'skipped'


class CAMExceptionType(Enum):
    """
    Exception type specified by the Agent. These categories are designed to be general enough to apply
    to various customers, payers, and claim processing scenarios.
    """
    # CAMs often run after EVAs, and some records might have failed processing at the eligibility level, perhaps missing benefits too.
    MISSING_ELIGIBILITY_OR_BENEFITS_INFO = 'missing_eligibility_or_benefits_info'
    
    # Related to missing information in the mapping file/store (e.g., value to correct is not listed in mapping).
    MAPPING_MISSING_OR_INCORRECT = 'mapping'
    
    # The claim to process for CAM was not in the right status, perhaps it was put on the queue in error.
    IMPROPER_STATUS = "improper_status"
    
    # An RPA error was encountered, could be a missing locator or some other RPA/browser related issue.
    RPA_ERROR = "rpa_error"
    
    # For desktop applications, when an error was encountered due to a missing locator, or someone else using the app (if the app is single user concurrently).
    DEKSTOP_ERROR = "desktop_error"
    
    # Something with a private API (i.e. not publicly documented) went wrong.
    PRIVATE_API = "private_api"
    
    # A file needed to complete the processing of a claim is missing, perhaps a file we/our customers are responsible for, or even an attachment like an x-ray or receipt.
    MISSING_REQUIRED_FILE = "missing_required_file"
    
    # Providers often miss or assign outdated / invalid CPT or diagnosis code
    INVALID_CPT_OR_ICD_CODE = "invalid_cpt_or_icd_code" 
    
    # Something about the state of the claim is wrong, like missing a modifier, missing a referring provider, missing a location associated or invalid units.
    IMPROPER_CLAIM_STATE = "improper_claim_state"

    # Claim had already been submitted / another claim with the same service and date of service already exists in the PMS.
    DUPLICATE_CLAIM = "duplicate_claim"

    # Provider may not be registered under a particular payer. (Often occurs with Medicare)
    INVALID_PROVIDER_NPI = "invalid_provider_npi" 
    
    # Related incomplete or invalid patient demographic data (e.g., missing gender or invalid zip code).
    PATIENT_DEMOGRAPHIC = 'patient_demographic'

    # Discrepancy where a patient's name does not match between different systems (e.g., Valant vs. Availity).
    PATIENT_NAME_MISMATCH = 'patient_name_mismatch'
    
    # Issues where the patient cannot be identified or matched in the system.
    PATIENT_IDENTIFICATION = 'patient_identification'
    
    # HTTP or connectivity-related issues (e.g., 404 Not Found, 401 Unauthorized) when accessing external APIs.
    HTTP_CONNECTIVITY = 'http_connectivity'
    
    # Internal validation or formatting issues in data fields (e.g., length restrictions not met).
    VALIDATION_FORMAT = 'validation_format'
    
    # Exceptions that do not fit well into the other categories. Typically, customer specific.
    # Before classifying an exception as "CUSTOM", consider if we need to create a new ENUM type (the current list
    # is not exhaustive and will be expanded over time to include new exception types that are generalizable across
    # customers.
    CUSTOM = 'custom'

    # A catch-all for unexpected or generic exceptions not covered by other categories.
    UNEXPECTED = 'unexpected'


class CAMRecord(BaseModel):
    status: CAMRecordStatus
    exception_type: Optional[CAMExceptionType]
    message: Optional[str]  # Additional context about the exception provided by the Agent.
    customer_id: str  # ID of Thoughtful customer
    claim_id: str  # ID of the claim being performed on
    work: str  # the work performed relating to the claim
    actioned: bool  # Whether the CAM agent did anything to the record at all, useful for CAMs where it is checking, and non-action is the ideal outcome 


from random import randint
from typing import Optional

from faker import Faker

from t_reporting_sdk.domain.agent_runs.models import AgentRun


class AgentRunFaker:
    @staticmethod
    def provide(
        run_id: Optional[int] = None,
        run_date: Optional[str] = None,
    ) -> AgentRun:
        faker = Faker()

        fake_run_id = randint(1, 100)
        fake_run_date = faker.date_time_this_month()

        return AgentRun(
            run_id=fake_run_id if run_id is None else run_id,
            run_date=fake_run_date if run_date is None else run_date,
        )


import os
import pytest

from t_reporting_sdk.config import ReportingSDKConfig


class TestReportingSDKConfig:
    def setup_class(self):
        os.environ["FABRIC_BASE_URL"] = "http://localhost:8000"
        os.environ["FABRIC_USER_EMAIL"] = "default@reporter.test"
        os.environ["FABRIC_USER_OTP_SECRET"] = "fake_secret"

    def teardown_class(self):
        del os.environ["FABRIC_BASE_URL"]
        del os.environ["FABRIC_USER_EMAIL"]
        del os.environ["FABRIC_USER_OTP_SECRET"]

    @pytest.mark.unit
    def test_singleton(self):
        config1 = ReportingSDKConfig()
        config2 = ReportingSDKConfig()
        config3 = ReportingSDKConfig.configure()

        assert config1 is config2
        assert config1 is config3

    @pytest.mark.unit
    def test_default_values(self):
        config = ReportingSDKConfig()

        assert config.fabric_client_config.base_url == "http://localhost:8000"
        assert config.fabric_client_config.user_email == "default@reporter.test"
        assert config.fabric_client_config.user_otp_secret == "fake_secret"

    @pytest.mark.unit
    def test_partial_configure(self):
        ReportingSDKConfig.configure(fabric_base_url="http://localhost:8001")

        config = ReportingSDKConfig()

        assert config.fabric_client_config.base_url == "http://localhost:8001"
        assert config.fabric_client_config.user_email == "default@reporter.test"
        assert config.fabric_client_config.user_otp_secret == "fake_secret"

    @pytest.mark.unit
    def test_configure(self):
        ReportingSDKConfig.configure(
            fabric_base_url="http://localhost:8001",
            fabric_user_email="non_default@reporter.test",
            fabric_user_secret="another_fake_secret",
        )

        config = ReportingSDKConfig()

        assert config.fabric_client_config.base_url == "http://localhost:8001"
        assert config.fabric_client_config.user_email == "non_default@reporter.test"
        assert config.fabric_client_config.user_otp_secret == "another_fake_secret"



from datetime import datetime, timezone
from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator


class PAULARecordStatus(Enum):
    # There was an exception or other error during the run
    ERROR = 'error'
    
    # We are skipping over authorization for this payer due to out of scope of automation or integration not yet implemented
    SKIPPED = 'skipped'
    
    # The authorization is still being reviewed
    PENDING = 'pending'
    
    # The authorization has been reviewed and was approved
    APPROVED = 'approved'
    
    # The authorization has been reviewed and was denied
    DENIED = 'denied'
    
    # The authorization is not required for the payer after the auth submission process
    NO_AUTH_REQUIRED = 'no_auth_required'


class PAULAExceptionType(Enum):
    """
    Exception type specified by the Agent. These categories are designed to be general enough to apply
    to various customers, payers, and portals.
    """

    # Related to missing information in the mapping file (e.g., payer not found in mapping file).
    MAPPING_FILE = 'mapping_file'

    # Related incomplete or invalid patient demographic data (e.g., missing gender or invalid zip code).
    PATIENT_DEMOGRAPHIC = 'patient_demographic'

    # Discrepancy where a patient’s name does not match between different systems (e.g., Valant vs. Availity).
    PATIENT_NAME_MISMATCH = 'patient_name_mismatch'

    # Exceptions related to retrieving payer details.
    NO_PAYER_FOUND = 'no_payer_found'

    # Situations involving multiple payers, extra COB info, or special secondary coverage complexities.
    MULTIPLE_PAYER_OR_COB = 'multiple_payer_or_cob'

    # Related to missing, invalid, or soon-to-expire plan effective/end dates.
    COVERAGE_DATE = 'coverage_date'

    # Issues where the patient cannot be identified or matched in the system.
    PATIENT_IDENTIFICATION = 'patient_identification'
    
    # Providers often miss or assign outdated / invalid CPT code
    INVALID_CPT_CODE = "invalid_cpt_code" 

    # Diagnosis codes could be missing or invalid
    INVALID_ICD_CODE = "invalid_icd_code"

    # The provider may put a number of units that exceeds the patients allowable limit
    INVALID_UNITS = "invalid_units"

    # Claim had already been submitted / another claim with the same service and date of service already exists in the PMS.
    DUPLICATE_AUTHORIZATION = "duplicate_authorization"

    # Provider may not be registered under a particular payer. (Often occurs with Medicare)
    INVALID_PROVIDER_NPI = "invalid_provider_npi" 

    # Some claims require a referring provider
    MISSING_REFERRING_PROVIDER = "missing_referring_provider"
    
    # Exception for missing required info (e.g., chart notes, medical docs, pain level, etc.).
    MISSING_MEDICAL_DOCUMENTATION = "missing_medical_documentation"

    # HTTP or connectivity-related issues (e.g., 404 Not Found, 401 Unauthorized) when accessing external APIs.
    HTTP_CONNECTIVITY = 'http_connectivity'

    # Internal validation or formatting issues in data fields (e.g., length restrictions not met).
    VALIDATION_FORMAT = 'validation_format'

    # Related to technical issues indicating unexpected code-level issues.
    TECHNICAL_ATTRIBUTE = 'technical_attribute'

    # Exceptions that do not fit well into the other categories. Typically, customer specific.
    # Before classifying an exception as "CUSTOM", consider if we need to create a new ENUM type (the current list
    # is not exhaustive and will be expanded over time to include new exception types that are generalizable across
    # customers.
    CUSTOM = 'custom'

    # A catch-all for unexpected or generic exceptions not covered by other categories.
    UNEXPECTED = 'unexpected'


class PAULARecordOriginalStatus(Enum):
    # Indicates that the authorization record is a standard submission request for prior authorization.
    NONE = 'none'
    
    # Indicates that the authorization record is a follow-up check on a previously submitted prior authorization.
    PENDING = 'pending'


class PAULARecord(BaseModel):
    model_config = ConfigDict(validate_assignment=True)

    status: PAULARecordStatus
    exception_type: Optional[PAULAExceptionType]
    message: Optional[str] # Additional context about the exception provided by the Agent.
    customer_id: str # ID of Thoughtful customer

    original_status: PAULARecordOriginalStatus
    patient_id: str
    portal: str
    appointment_date: datetime
    treatment_code: str # HCPC, CPT, procedure
    payer: Optional[str] # required for PENDING, DENIED, APPROVED
    payer_id: Optional[str] # required for PENDING, DENIED, APPROVED
    plan_id: Optional[str] # Required if we have a payer/payer_id
    system_of_record_id: str # ID of the underyling system assoicated with the prior auth
    authorization_id: Optional[str] # required for DENIED, APPROVED. Refers to reference ID

    @field_validator('appointment_date', mode='before')
    @classmethod
    def ensure_timezone(cls, value: datetime) -> datetime:
        if value.tzinfo is None:
            raise ValueError('appointment_date must be timezone-aware')
        return value.astimezone(timezone.utc)
    
    @field_validator('payer', 'payer_id', mode='before')
    @classmethod
    def ensure_payer(cls, value: Optional[str], info: ValidationInfo) -> Optional[str]:
        if info.data['status'] in [PAULARecordStatus.PENDING, PAULARecordStatus.DENIED, PAULARecordStatus.APPROVED]:
            if value is None:
                raise ValueError('payer and payer_id are required for PENDING, DENIED, and APPROVED status')
        return value
    
    @field_validator('plan_id', mode='before')
    @classmethod
    def ensure_plan_id(cls, value: Optional[str], info: ValidationInfo) -> Optional[str]:
        if info.data['payer'] is not None and value is None:
            raise ValueError('plan_id is required when payer is provided')
        return value
    
    @field_validator('authorization_id', mode='before')
    @classmethod
    def ensure_authorization_id(cls, value: Optional[str], info: ValidationInfo) -> Optional[str]:
        if info.data['status'] in [PAULARecordStatus.DENIED, PAULARecordStatus.APPROVED]:
            if value is None:
                raise ValueError('authorization_id is required for DENIED and APPROVED status')
        return value



from enum import Enum
from typing import Optional

from pydantic import BaseModel


class EVARecordStatus(Enum):
    # Successfully conducted eligibility check. No issues/warnings encountered.
    SUCCESS = 'success'

    # Successfully conducted eligibility check but encountered a scenario that requires staff review.
    WARNING = 'warning'

    # Unable to conduct eligibility check
    ERROR = 'error'

    # Skipped typically indicates that the scenario is a known-out of scope scenario
    SKIPPED = 'skipped'


class EVAExceptionType(Enum):
    """
    Exception type specified by the Agent. These categories are designed to be general enough to apply
    to various customers, payers, and portals.
    """
    # Related to missing information in the mapping file (e.g., payer not found in mapping file).
    MAPPING_FILE = 'mapping_file'

    # Related incomplete or invalid patient demographic data (e.g., missing gender or invalid zip code).
    PATIENT_DEMOGRAPHIC = 'patient_demographic'

    # Discrepancy where a patient’s name does not match between different systems (e.g., Valant vs. Availity).
    PATIENT_NAME_MISMATCH = 'patient_name_mismatch'

    # Exceptions related to retrieving coverage details from external portals or payers.
    COVERAGE_RETRIEVAL = 'coverage_retrieval'

    # Exceptions related to retrieving payer details.
    NO_PAYER_FOUND = 'no_payer_found'

    # Situations involving multiple payers, extra COB info, or special secondary coverage complexities.
    MULTIPLE_PAYER_OR_COB = 'multiple_payer_or_cob'

    # Related to missing, invalid, or soon-to-expire plan effective/end dates.
    COVERAGE_DATE = 'coverage_date'

    # Copay-related issues, such as unusually large copays that require review.
    COPAY = 'copay'

    # Issues where the patient cannot be identified or matched in the system.
    PATIENT_IDENTIFICATION = 'patient_identification'

    # HTTP or connectivity-related issues (e.g., 404 Not Found, 401 Unauthorized) when accessing external APIs.
    HTTP_CONNECTIVITY = 'http_connectivity'

    # Internal validation or formatting issues in data fields (e.g., length restrictions not met).
    VALIDATION_FORMAT = 'validation_format'

    # Related to technical issues indicating unexpected code-level issues.
    TECHNICAL_ATTRIBUTE = 'technical_attribute'

    # Exceptions that do not fit well into the other categories. Typically, customer specific.
    # Before classifying an exception as "CUSTOM", consider if we need to create a new ENUM type (the current list
    # is not exhaustive and will be expanded over time to include new exception types that are generalizable across
    # customers.
    CUSTOM = 'custom'

    # A catch-all for unexpected or generic exceptions not covered by other categories.
    UNEXPECTED = 'unexpected'


class EVAInsuranceEligibility(Enum):
    ELIGIBLE = 'eligible'
    INELIGIBLE = 'ineligible'


class EVAInsuranceType(Enum):
    PRIMARY = 'primary'
    SECONDARY = 'secondary'
    TERTIARY = 'tertiary'
    QUATERNARY = 'quaternary'


class EVARecord(BaseModel):
    status: EVARecordStatus
    exception_type: Optional[EVAExceptionType]
    message: str  # Additional context about the exception provided by the Agent.
    customer_id: str  # ID of the Thoughtful customer

    # EVA records typically refer to a patient/payer combo run on a specific portal
    patient_id: str  # ID of the patient eligibility verification is being performed on
    payer_id: Optional[str]   # ID of the Payer that eligibility verification is being checked against
    payer_name: Optional[str] = None # Name of the payer that eligibility verification is being checked against
    portal: str  # Name of the portal being interacted with
    insurance_eligibility: Optional[EVAInsuranceEligibility] = None # Insurance eligibility status of the patient
    insurance_type: Optional[EVAInsuranceType] = None  # Type of insurance (primary, secondary, etc.)



from t_reporting_sdk.domain.agent_runs.models import AgentRun
from t_reporting_sdk.domain.paula_records.models import PAULARecord
from t_reporting_sdk.repositories.api_clients.fabric.client import FabricClient


class PAULARecordsRepository:
    def __init__(self, client: FabricClient):
        self._client = client

    def store_paula_record(
        self,
        agent_run: AgentRun,  # AgentRun to associate the PAULARecord with
        paula_record: PAULARecord,
    ) -> None:
        self._client.create_paula_record(
            run_id=agent_run.run_id,
            status=paula_record.status.value,
            exception_type=None if paula_record.exception_type is None else paula_record.exception_type.value,
            message=paula_record.message,
            customer_id=paula_record.customer_id,
            original_status=None if paula_record.original_status is None else paula_record.original_status.value,
            patient_id=paula_record.patient_id,
            portal=paula_record.portal,
            appointment_date=paula_record.appointment_date.isoformat(),
            treatment_code=paula_record.treatment_code,
            payer=paula_record.payer,
            payer_id=paula_record.payer_id,
            plan_id=paula_record.plan_id,
            system_of_record_id=paula_record.system_of_record_id,
            authorization_id=paula_record.authorization_id,
        )



import os
from typing import Optional

from t_reporting_sdk.repositories.api_clients.fabric.client import FabricClientConfig


class ReportingSDKConfig:
    """Singleton configuration class for the SDK."""
    _instance: Optional["ReportingSDKConfig"] = None

    fabric_client_config: FabricClientConfig

    def __new__(cls) -> "ReportingSDKConfig":
        """Singleton instance creation."""
        if cls._instance is None:
            # The super() function here refers to the base (object) class, which is the default parent of all Python classes. 
            # Calling super().__new__(cls) invokes the object class's __new__ method to allocate memory for a new instance. 
            # This is crucial for the singleton pattern, as it ensures proper instance creation and allows us to enforce 
            # that only one instance of the class is ever created. The intent of this code is to control object creation 
            # and reuse a single shared instance across the application.
            cls._instance = super().__new__(cls)
            # This ensures that the config is always initialized with values
            cls._instance.configure()
        return cls._instance

    @classmethod
    def configure(
        cls,
        *,
        fabric_base_url: Optional[str] = None,
        fabric_user_email: Optional[str] = None,
        fabric_user_secret: Optional[str] = None,
    ) -> "ReportingSDKConfig":
        """Set the configuration for the SDK."""
        if cls._instance is None:
            cls._instance = super().__new__(cls)

        cls._instance.fabric_client_config = FabricClientConfig(
            base_url=fabric_base_url or os.getenv("FABRIC_BASE_URL"),
            user_email=fabric_user_email or os.getenv("FABRIC_USER_EMAIL"),
            user_otp_secret=fabric_user_secret or os.getenv("FABRIC_USER_OTP_SECRET"),
        )

        return cls._instance



import random
from random import randint
from typing import Optional

from faker import Faker

from t_reporting_sdk.domain.cam_records.models import CAMRecord, CAMRecordStatus, CAMExceptionType


class CAMRecordFaker:
    """
    Generates fake CAMRecord objects for testing purposes.
    """

    @staticmethod
    def provide(
            status: Optional[CAMRecordStatus] = None,
            exception_type: Optional[CAMExceptionType] = None,
            message: Optional[str] = None,
            customer_id: Optional[str] = None,
            claim_id: Optional[str] = None,
            work: Optional[str] = None,
            actioned: Optional[bool] = None,
    ) -> CAMRecord:
        """
        Provides a fake CAMRecord with the given parameters, or random values if not provided.
        """
        faker = Faker()
        
        fake_status = random.choice(list(CAMRecordStatus))
        fake_exception_type = random.choice(list(CAMExceptionType)) if fake_status in [CAMRecordStatus.ERROR, CAMRecordStatus.WARNING] else None
        
        if fake_exception_type is not None:
            fake_message = f"Encountered {fake_exception_type.value} issue: {faker.sentence()}"
        else:
            fake_message = faker.sentence() if fake_status != CAMRecordStatus.SUCCESS else ""
            
        fake_customer_id = str(randint(1, 100))
        fake_claim_id = str(randint(1, 100))
        fake_work = faker.word()
        fake_actioned = random.choice([True, False])
        
        # If status is not ERROR or WARNING, exception_type should be None
        if status is not None and status not in [CAMRecordStatus.ERROR, CAMRecordStatus.WARNING]:
            exception_type = None

        return CAMRecord(
            status=status or fake_status,
            exception_type=exception_type or fake_exception_type,
            message=message or fake_message,
            customer_id=customer_id or fake_customer_id,
            claim_id=claim_id or fake_claim_id,
            work=work or fake_work,
            actioned=actioned or fake_actioned,
        ) 


from t_reporting_sdk.repositories.agent_runs.repository import AgentRunsRepository
from t_reporting_sdk.repositories.api_clients.fabric.client import FabricClient
from t_reporting_sdk.repositories.eva_records.repository import EVARecordsRepository
from t_reporting_sdk.repositories.cam_records.repository import CAMRecordsRepository
from t_reporting_sdk.repositories.paula_records.repository import PAULARecordsRepository
from t_reporting_sdk.config import ReportingSDKConfig

sdk_config = ReportingSDKConfig()
fabric_client = FabricClient(config=sdk_config.fabric_client_config)
default_agent_runs_repository = AgentRunsRepository(client=fabric_client)
default_eva_records_repository = EVARecordsRepository(client=fabric_client)
default_cam_records_repository = CAMRecordsRepository(client=fabric_client)
default_paula_records_repository = PAULARecordsRepository(client=fabric_client)



from concurrent.futures import ThreadPoolExecutor, as_completed

import logging
from typing import Optional, List

from t_reporting_sdk.domain.agent_runs.models import AgentRun
from t_reporting_sdk.domain.eva_records.models import EVARecord
from t_reporting_sdk.repositories.agent_runs.repository import AgentRunsRepository
from t_reporting_sdk.repositories.eva_records.repository import EVARecordsRepository


def _load_default_eva_records_repository() -> EVARecordsRepository:
    """
    Handles dependency injection automatically so that SDK users do not have to specify their dependencies manually.
    Handled as a separate function so that tests do not try to import the default repositories
    """
    from t_reporting_sdk.default_dependencies import default_eva_records_repository
    return default_eva_records_repository


def _load_default_agent_runs_repository() -> AgentRunsRepository:
    """
    Handles dependency injection automatically so that SDK users do not have to specify their dependencies manually.
    Handled as a separate function so that tests do not try to import the default repositories
    """
    from t_reporting_sdk.default_dependencies import default_agent_runs_repository
    return default_agent_runs_repository


class EVAReporter:
    def __init__(
            self,
            agent_run: AgentRun,
            eva_repository: Optional[EVARecordsRepository] = None,
            agent_runs_repository: Optional[AgentRunsRepository] = None,
    ):
        # Load default repositories if none are provided
        self._eva_repository = _load_default_eva_records_repository() if eva_repository is None else eva_repository
        agent_runs_repository = _load_default_agent_runs_repository() if agent_runs_repository is None else agent_runs_repository

        self._agent_run = agent_run

        # Ensure Agent Run is stored in the database so that we can associate EVA results with it
        agent_runs_repository.store_agent_run(agent_run=agent_run)

    def report_eva_records(
            self,
            eva_records: List[EVARecord],
            # Limit the # of saves made parallel - saves will fail if # of parallel calls > connection_pool size
            max_records_to_store_in_parallel: int = 4,
    ):
        def store_record(eva_record: EVARecord):
            try:
                self._eva_repository.store_eva_record(
                    agent_run=self._agent_run,
                    eva_record=eva_record,
                )
            except Exception as e:
                logging.error(f"Unable to report eligibility verification result: {e}")

        with ThreadPoolExecutor(max_workers=max_records_to_store_in_parallel) as executor:
            futures = {executor.submit(store_record, record): record for record in eva_records}

            for future in as_completed(futures):
                record = futures[future]
                try:
                    future.result()  # Will raise an exception if the task failed
                except Exception as e:
                    logging.error(f"Failed to process record {record}: {e}")


class EvaReporter(EVAReporter):
    def __new__(cls, *args, **kwargs):
        logging.warning(f"{cls.__name__} is deprecated. Please use EVAReporter instead.")
        return super().__new__(cls)


from concurrent.futures import ThreadPoolExecutor, as_completed

import logging
from typing import Optional, List

from t_reporting_sdk.domain.agent_runs.models import AgentRun
from t_reporting_sdk.domain.paula_records.models import PAULARecord
from t_reporting_sdk.repositories.agent_runs.repository import AgentRunsRepository
from t_reporting_sdk.repositories.paula_records.repository import PAULARecordsRepository


def _load_default_paula_records_repository() -> PAULARecordsRepository:
    """
    Handles dependency injection automatically so that SDK users do not have to specify their dependencies manually.
    Handled as a separate function so that tests do not try to import the default repositories
    """
    from t_reporting_sdk.default_dependencies import default_paula_records_repository
    return default_paula_records_repository


def _load_default_agent_runs_repository() -> AgentRunsRepository:
    """
    Handles dependency injection automatically so that SDK users do not have to specify their dependencies manually.
    Handled as a separate function so that tests do not try to import the default repositories
    """
    from t_reporting_sdk.default_dependencies import default_agent_runs_repository
    return default_agent_runs_repository


class PAULAReporter:
    def __init__(
        self,
        agent_run: AgentRun,
        paula_repository: Optional[PAULARecordsRepository] = None,
        agent_runs_repository: Optional[AgentRunsRepository] = None,
    ):
        # Load default repositories if none are provided
        self._paula_repository = paula_repository or _load_default_paula_records_repository()
        agent_runs_repository = agent_runs_repository or _load_default_agent_runs_repository()

        self._agent_run = agent_run

        # Ensure Agent Run is stored in the database so that we can associate EVA results with it
        agent_runs_repository.store_agent_run(agent_run=agent_run)

    def report_paula_records(
        self,
        paula_records: List[PAULARecord],
        # Limit the # of saves made parallel - saves will fail if # of parallel calls > connection_pool size
        max_records_to_store_in_parallel: int = 4,
    ):
        def store_record(paula_record: PAULARecord):
            from requests.exceptions import HTTPError
            try:
                self._paula_repository.store_paula_record(
                    agent_run=self._agent_run,
                    paula_record=paula_record,
                )
            except HTTPError as e:
                logging.exception(f"Failed to store record")
                logging.error(e.response.text)
            except Exception as e:
                logging.exception(f"Failed to store record")
                logging.error(f"Unable to report prior authorization result: {e}")

        with ThreadPoolExecutor(max_workers=max_records_to_store_in_parallel) as executor:
            futures = {executor.submit(store_record, record): record for record in paula_records}

            for future in as_completed(futures):
                record = futures[future]
                try:
                    future.result()
                except Exception as e:
                    logging.error(f"Failed to process record: {record}: {e}")



from t_reporting_sdk.domain.agent_runs.models import AgentRun
from t_reporting_sdk.domain.cam_records.models import CAMRecord
from t_reporting_sdk.repositories.api_clients.fabric.client import FabricClient


class CAMRecordsRepository:
    def __init__(self, client: FabricClient):
        self._client = client

    def store_cam_record(
            self,
            agent_run: AgentRun,  # AgentRun to associate the CAMRecord with
            cam_record: CAMRecord,
    ) -> None:
        self._client.create_cam_record(
            run_id=agent_run.run_id,
            status=cam_record.status.value,
            exception_type=None if cam_record.exception_type is None else cam_record.exception_type.value,
            message=cam_record.message,
            customer_id=cam_record.customer_id,
            claim_id=cam_record.claim_id,
            work=cam_record.work,
            actioned=cam_record.actioned,
        ) 